# -*- coding: utf-8 -*-
import asyncio
import ssl
from asyncio import IncompleteReadError
from unittest import mock
from unittest.mock import patch, MagicMock

import pytest
from aioamqp.protocol import CLOSED, CLOSING, CONNECTING, OPEN

import cabbagok
from cabbagok import AmqpConnection
from cabbagok.amqp import aioamqp_connect
from tests.conftest import MockProtocol, MockTransport

pytestmark = pytest.mark.asyncio

# Sample non-default connection info:
HOST = 'fake_amqp_host'
PORT = 25762
USERNAME = 'amqp_user'
PASSWORD = 'secret_password'
VIRTUALHOST = 'test_vhost'


@pytest.mark.asyncio
async def test_amqp_connection_ssl(event_loop):
    ssl_context = mock.MagicMock()
    mock_transport, mock_protocol = MockTransport(), MockProtocol()
    with patch('asyncio.base_events.BaseEventLoop.create_connection') as mock_connect:
        mock_connect.return_value = (mock_transport, mock_protocol)
        await aioamqp_connect(host=HOST, port=PORT, virtualhost=VIRTUALHOST, login=USERNAME, password=PASSWORD,
                              loop=event_loop, ssl=ssl_context)
        mock_connect.assert_called_once_with(mock.ANY, HOST, PORT, ssl=ssl_context)


@pytest.mark.asyncio
async def test_amqp_connection_exception(event_loop):
    ssl_context = mock.MagicMock()
    mock_transport, mock_protocol = MockTransport(), MockProtocol()
    mock_protocol.start_connection = mock.MagicMock(side_effect=ValueError('TestException'))
    with patch('asyncio.base_events.BaseEventLoop.create_connection') as mock_connect:
        mock_connect.return_value = (mock_transport, mock_protocol)
        with pytest.raises(ValueError):
            await aioamqp_connect(host=HOST, port=PORT, virtualhost=VIRTUALHOST, login=USERNAME, password=PASSWORD,
                                  loop=event_loop, ssl=ssl_context)


@pytest.mark.asyncio
async def test_amqp_connection_ssl_context(event_loop):
    mock_transport, mock_protocol = MockTransport(), MockProtocol()
    with patch('asyncio.base_events.BaseEventLoop.create_connection') as mock_connect:
        with patch('ssl.create_default_context') as mock_default_context:
            mock_connect.return_value = (mock_transport, mock_protocol)
            await aioamqp_connect(host=HOST, port=PORT, virtualhost=VIRTUALHOST, login=USERNAME, password=PASSWORD,
                                  verify_ssl=False, ssl=mock_default_context)
            assert mock_default_context.check_hostname is False
            assert mock_default_context.verify_mode == ssl.CERT_NONE


@pytest.mark.asyncio
async def test_amqp_connection_with_ssl_port(event_loop):
    SSL_PORT = 5671
    mock_transport, mock_protocol = MockTransport(), MockProtocol()
    with patch('asyncio.base_events.BaseEventLoop.create_connection') as mock_connect:
        mock_connect.return_value = (mock_transport, mock_protocol)
        await aioamqp_connect(host=HOST, virtualhost=VIRTUALHOST, login=USERNAME, password=PASSWORD, loop=event_loop,
                              ssl=True)
        mock_connect.assert_called_once_with(mock.ANY, HOST, SSL_PORT, ssl=mock.ANY)


@pytest.mark.asyncio
async def test_amqp_connection_with_default_port(event_loop):
    DEFAULT_PORT = 5672
    mock_transport, mock_protocol = MockTransport(), MockProtocol()
    with patch('asyncio.base_events.BaseEventLoop.create_connection') as mock_connect:
        mock_connect.return_value = (mock_transport, mock_protocol)
        await aioamqp_connect(host=HOST, virtualhost=VIRTUALHOST, login=USERNAME, password=PASSWORD, loop=event_loop,
                              ssl=False)
        mock_connect.assert_called_once_with(mock.ANY, HOST, DEFAULT_PORT)


class TestConnect:
    """AmqpConnection.connect"""

    async def test_connect(self, event_loop):
        """Check typical connection call."""
        mock_transport, mock_protocol = MockTransport(), MockProtocol()
        connection = cabbagok.AmqpConnection(hosts=[(HOST, PORT)], username=USERNAME, password=PASSWORD,
                                             virtualhost=VIRTUALHOST, loop=event_loop)
        with patch('cabbagok.amqp.aioamqp_connect') as mock_connect:
            mock_connect.return_value = (mock_transport, mock_protocol)
            await connection.connect()

        mock_connect.assert_called_once_with(host=HOST, port=PORT, login=USERNAME, password=PASSWORD,
                                             virtualhost=VIRTUALHOST, loop=event_loop, ssl=False)
        assert connection.transport is mock_transport
        assert connection.protocol is mock_protocol

    async def test_default_params(self, event_loop):
        """Check that AmqpConnection supplies reasonable defaults."""
        connection = cabbagok.AmqpConnection(loop=event_loop)
        mock_transport, mock_protocol = MockTransport(), MockProtocol()
        with patch('cabbagok.amqp.aioamqp_connect') as mock_connect:
            mock_connect.return_value = (mock_transport, mock_protocol)
            await connection.connect()

        mock_connect.assert_called_once_with(host='localhost', port=5672, login='guest', password='guest',
                                             virtualhost='/', loop=event_loop, ssl=False)
        assert connection.transport is mock_transport
        assert connection.protocol is mock_protocol

    async def test_called_twice(self, event_loop):
        """Check that calling the method twice has no effect."""
        connection = cabbagok.AmqpConnection(loop=event_loop)
        mock_transport, mock_protocol = MockTransport(), MockProtocol()
        connection.transport = mock_transport
        connection.protocol = mock_protocol
        with patch('cabbagok.amqp.aioamqp_connect') as mock_connect:
            await connection.connect()

        mock_connect.assert_not_called()
        assert connection.transport is mock_transport
        assert connection.protocol is mock_protocol

    async def test_retry(self, event_loop):
        """Keep reconnecting if an error occurs."""
        attempts = 10

        async def faulty_connect(_attempts=[attempts], **kwargs):
            assert kwargs == dict(host=HOST, port=PORT, login=USERNAME, password=PASSWORD, virtualhost=VIRTUALHOST,
                                  loop=event_loop, ssl=False)
            _attempts[0] -= 1
            if _attempts[0]:
                raise OSError('[Errno 113] Connect call failed')
            return mock_transport, mock_protocol

        connection = cabbagok.AmqpConnection(hosts=[(HOST, PORT)], username=USERNAME,
                                             password=PASSWORD, virtualhost=VIRTUALHOST, loop=event_loop)
        mock_transport, mock_protocol = MockTransport(), MockProtocol()

        with patch('cabbagok.amqp.aioamqp_connect', new=faulty_connect), patch('asyncio.sleep') as mock_sleep:
            await connection.connect()

        assert mock_sleep.call_count == attempts - 1
        assert connection.transport is mock_transport
        assert connection.protocol is mock_protocol

    async def test_fatal_error(self, event_loop):
        """Some connection errors are not worth trying to recover from."""
        connection = cabbagok.AmqpConnection(hosts=[('angrydev.ru', 80)], loop=event_loop)

        with pytest.raises(IncompleteReadError):
            with patch('cabbagok.amqp.aioamqp_connect',
                       side_effect=IncompleteReadError([], 160)) as mock_connect:
                await connection.connect()

        mock_connect.assert_called_once()
        assert connection.transport is None
        assert connection.protocol is None


class TestDisconnect:
    async def test_ok(self, connection):
        await connection.disconnect()
        connection.protocol.close.assert_called_once_with()

    async def test_no_protocol(self, event_loop):
        connection = cabbagok.AmqpConnection(loop=event_loop)
        connection.protocol = MagicMock()
        await connection.disconnect()
        connection.protocol.close.assert_not_called()

    async def test_not_connected(self, connection):
        connection.protocol.state = CLOSED
        await connection.disconnect()
        connection.protocol.close.assert_not_called()


class TestIsConnectedProperty:
    async def test_protocol_is_none(self, connection):
        connection.protocol = None
        assert connection.is_connected is False

    @pytest.mark.parametrize('state', [CONNECTING, CLOSING, CLOSED])
    async def test_protocol_state_not_equals_open(self, connection: AmqpConnection, state):
        connection.protocol.state = state
        assert connection.is_connected is False

    async def test_ok(self, connection: AmqpConnection):
        connection.protocol = MockProtocol()
        connection.protocol.state = OPEN
        assert connection.is_connected is True



# -*- coding: utf-8 -*-
class ExponentialBackoff:
    """
    Exponential backoff with an upper limit.

    >>> delay = ExponentialBackoff(1, 2, 100)
    >>> [delay.next() for _ in range(10)]
    [1, 2, 4, 8, 16, 32, 64, 100, 100, 100]

    """

    def __init__(self, start, factor, limit):
        self.current = start
        self.factor = factor
        self.limit = limit

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self) -> int:
        try:
            return round(self.current)
        finally:
            self.current = min(self.current * self.factor, self.limit)


class FibonaccianBackoff(ExponentialBackoff):
    """
    Exponential backoff with parameters that produce a Fibonacci-like sequence:

    1, 1, 2, 3, 5, 8, 13, 21, ...
    """

    def __init__(self, limit):
        super().__init__(start=0.724, factor=1.618, limit=limit)



# -*- coding: utf-8 -*-
from unittest.mock import patch, MagicMock

import aioamqp
import asyncio
import random
from itertools import islice

import pytest

import cabbagok
from tests.conftest import MockTransport, MockProtocol, SUBSCRIPTION_QUEUE, TEST_EXCHANGE, SUBSCRIPTION_KEY, \
    RANDOM_QUEUE, HOST, MockEnvelope, MockProperties, CONSUMER_TAG, DELIVERY_TAG, RESPONSE_CORR_ID

pytestmark = pytest.mark.asyncio
TEST_DELAY = 0.2


class TestConnect:
    """AsyncAmqpRpc.connect"""

    @pytest.mark.parametrize('exchange', ['', 'seldom-exchange', 'public', 'private'])
    async def test_ok(self, event_loop, exchange):
        connection = cabbagok.AmqpConnection(hosts=[(HOST, 5672)], loop=event_loop)
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        rpc.callback_exchange = exchange
        with patch('cabbagok.amqp.aioamqp_connect') as mock_connect:
            mock_connect.return_value = (MockTransport(), MockProtocol())
            await rpc.connect()
        mock_connect.assert_called_once()
        # check that client is set up correctly:
        assert isinstance(rpc.channel, aioamqp.channel.Channel)  # actually it's a mock pretending to be a Channel
        rpc.channel.queue_declare.assert_called_with(exclusive=True)
        rpc.channel.basic_consume.assert_called_once_with(callback=rpc._on_response,
                                                          queue_name=rpc.callback_queue)
        if rpc.callback_exchange != '':
            rpc.channel.exchange_declare.assert_called_once_with(
                exchange_name=rpc.callback_exchange, type_name='topic', durable=True)
            rpc.channel.queue_bind.assert_called_once_with(queue_name=rpc.callback_queue,
                                                           exchange_name=rpc.callback_exchange,
                                                           routing_key=rpc.callback_queue)
        else:
            rpc.channel.exchange_declare.assert_not_called()
            rpc.channel.queue_bind.assert_not_called()


class TestSubscribe:
    """AsyncAmqpRpc.subscribe"""

    async def test_ok(self, connection):
        def request_handler(request):
            return request

        rpc = cabbagok.AsyncAmqpRpc(
            connection=connection,
            queue_params=dict(passive=False, durable=True, exclusive=True, auto_delete=True),
            exchange_params=dict(type_name='fanout', passive=False, durable=True, auto_delete=True))
        await rpc.connect()
        assert rpc.channel.queue_declare.call_count == 1
        assert rpc.channel.basic_consume.call_count == 1

        await rpc.subscribe(exchange=TEST_EXCHANGE, queue=SUBSCRIPTION_QUEUE, routing_key=SUBSCRIPTION_KEY,
                            request_handler=request_handler)
        assert rpc.channel.queue_declare.call_count == 2
        assert rpc.channel.basic_consume.call_count == 2
        rpc.channel.basic_qos.assert_called_once_with(
            prefetch_count=rpc.prefetch_count, prefetch_size=0, connection_global=False)
        rpc.channel.exchange_declare.assert_called_once_with(
            exchange_name=TEST_EXCHANGE, type_name='fanout', passive=False, durable=True, auto_delete=True)
        rpc.channel.queue_bind.assert_called_once_with(
            exchange_name=TEST_EXCHANGE, queue_name=SUBSCRIPTION_QUEUE, routing_key=SUBSCRIPTION_KEY)
        rpc.channel.queue_declare.assert_called_with(
            queue_name=SUBSCRIPTION_QUEUE, auto_delete=True, durable=True, exclusive=True, passive=False)
        # rpc.channel.basic_consume.assert_called_with(  # partial() != partial()
        #     callback=partial(rpc._on_request, request_handler=request_handler), queue_name=SUBSCRIPTION_QUEUE)

    async def test_defaults(self, connection):
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        assert rpc.channel.queue_declare.call_count == 1
        assert rpc.channel.basic_consume.call_count == 1

        await rpc.subscribe(request_handler=lambda x: x, queue=SUBSCRIPTION_QUEUE)
        assert rpc.channel.queue_declare.call_count == 2
        assert rpc.channel.basic_consume.call_count == 2
        rpc.channel.basic_qos.assert_called_once_with(
            prefetch_count=rpc.prefetch_count, prefetch_size=0, connection_global=False)
        rpc.channel.queue_declare.assert_called_with(
            queue_name=SUBSCRIPTION_QUEUE, durable=True, arguments={'x-dead-letter-exchange': 'DLX',
                                                                    'x-dead-letter-routing-key': 'dlx_rpc'})
        # rpc.channel.basic_consume.assert_called_with(
        #     callback=rpc._on_request, queue_name=SUBSCRIPTION_QUEUE)
        # It is an error to attempt declaring or binding on default exchange:
        rpc.channel.exchange_declare.assert_not_called()
        rpc.channel.queue_bind.assert_not_called()

    async def test_amqp_defaults(self, connection):
        """Test that broker defaults (empty queue, empty exchange) are handled well."""
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        assert rpc.channel.queue_declare.call_count == 1
        assert rpc.channel.basic_consume.call_count == 1

        await rpc.subscribe(request_handler=lambda x: x, exchange='', queue='')
        assert rpc.channel.queue_declare.call_count == 2
        assert rpc.channel.basic_consume.call_count == 2
        rpc.channel.basic_qos.assert_called_once_with(
            prefetch_count=rpc.prefetch_count, prefetch_size=0, connection_global=False)
        rpc.channel.queue_declare.assert_called_with(
            queue_name='', durable=True, arguments={'x-dead-letter-exchange': 'DLX',
                                                    'x-dead-letter-routing-key': 'dlx_rpc'})
        # rpc.channel.basic_consume.assert_called_with(
        #     callback=rpc._on_request, queue_name=RANDOM_QUEUE)
        # We are still on default exchange:
        rpc.channel.exchange_declare.assert_not_called()
        rpc.channel.queue_bind.assert_not_called()


class TestUnsubscribe:
    """AsyncAmqpRpc.unsubscribe"""

    async def test_ok(self, connection):
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        await rpc.unsubscribe(consumer_tag=CONSUMER_TAG)
        rpc.channel.basic_cancel.assert_called_once_with(consumer_tag=CONSUMER_TAG)


class TestHandleRpc:
    """AsyncAmqpRpc.handle_rpc: low-level handler called by aioamqp"""

    @staticmethod
    def request_handler_factory(async_, responding=True, fail=False):
        """Creates a request_handler the server can call on the payload."""
        if fail:
            return MagicMock(side_effect=Exception('Handler error'))
        if async_:
            async def request_handler(request):
                return request if responding else None
        else:
            def request_handler(request):
                return request if responding else None
        return MagicMock(side_effect=request_handler)

    @pytest.mark.parametrize('is_async', [True, False])
    @pytest.mark.parametrize('body, expected', [
        (b'', ''),
        (b'Test message body. \xd0\xa2\xd0\xb5\xd1\x81\xd1\x82', 'Test message body. Тест'),
    ])
    async def test_responding(self, connection, body, expected, is_async):
        handler = self.request_handler_factory(is_async, responding=True, fail=False)
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        await rpc.handle_rpc(channel=rpc.channel, body=body, envelope=MockEnvelope(), properties=MockProperties(),
                             request_handler=handler)
        handler.assert_called_once_with(expected)
        rpc.channel.basic_client_nack.assert_not_called()
        rpc.channel.basic_client_ack.assert_called_once_with(delivery_tag=DELIVERY_TAG)
        rpc.channel.basic_publish.assert_called_once_with(
            exchange_name='', payload=body, properties={'correlation_id': RESPONSE_CORR_ID}, routing_key=RANDOM_QUEUE)

    @pytest.mark.parametrize('is_async', [True, False])
    @pytest.mark.parametrize('body, expected', [
        (b'', ''),
        (b'Test message body. \xd0\xa2\xd0\xb5\xd1\x81\xd1\x82', 'Test message body. Тест'),
    ])
    async def test_not_responding(self, connection, body, expected, is_async):
        """Handler returns None instead of str/bytes => no response needed."""
        handler = self.request_handler_factory(is_async, responding=False, fail=False)
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        await rpc.handle_rpc(channel=rpc.channel, body=body, envelope=MockEnvelope(), properties=MockProperties(),
                             request_handler=handler)
        handler.assert_called_once_with(expected)
        rpc.channel.basic_client_nack.assert_not_called()
        rpc.channel.basic_client_ack.assert_called_once_with(delivery_tag=DELIVERY_TAG)
        rpc.channel.basic_publish.assert_not_called()

    @pytest.mark.parametrize('is_async', [True, False])
    @pytest.mark.parametrize('body, expected', [
        (b'', ''),
        (b'Test message body. \xd0\xa2\xd0\xb5\xd1\x81\xd1\x82', 'Test message body. Тест'),
    ])
    async def test_exception(self, connection, body, expected, is_async):
        handler = self.request_handler_factory(is_async, fail=True)
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        await rpc.handle_rpc(channel=rpc.channel, body=body, envelope=MockEnvelope(), properties=MockProperties(),
                             request_handler=handler)
        handler.assert_called_once_with(expected)
        rpc.channel.basic_client_nack.assert_called_once_with(delivery_tag=DELIVERY_TAG)
        rpc.channel.basic_client_ack.assert_not_called()

    @pytest.mark.parametrize('is_connected', [True, False])
    @pytest.mark.parametrize('channel', [True, False])
    async def test_wait_connection(self, is_connected, channel):
        """
        Checking for execution time of cabbagok.AsyncAmqpRpc.wait_connected()
        In case of True value of both variables 'connection.is_connected' and 'channel'
        the function should terminate immediately. Else the function should check
        the state of these variables with an interval determined in variable 'connection_delay' (in seconds)
        until both variables won't have True value.
        """

        class FakeSelf:
            class FakeConnection:
                def __init__(self, is_connected_):
                    self.is_connected = is_connected_

            def __init__(self, is_connected_, channel_):
                self.connection = self.FakeConnection(is_connected_)
                self.channel = channel_
                self.connection_delay = TEST_DELAY

        fake_self = FakeSelf(is_connected, channel)
        future = asyncio.ensure_future(cabbagok.AsyncAmqpRpc.wait_connected(fake_self))
        await asyncio.sleep(TEST_DELAY)
        if not future.done():
            fake_self.connection.is_connected = True
            fake_self.channel = True

        await asyncio.sleep(TEST_DELAY)
        assert future.done()

    @pytest.mark.parametrize('number_of_tasks', [0, 1, 10])
    @pytest.mark.parametrize('pending', [True, False])
    async def test_launch_server(self, connection, number_of_tasks, pending):
        """
        Test for cabbagok.AsyncAmqpRpc.stop(). All tasks should execute asynchronously.
        In process of tests it may be created a big task (pending variable) if compare with others ones.
        In this case the task should continue the executing after calling the target function
        """

        small_delay = TEST_DELAY * 0.5
        big_delay = TEST_DELAY * 3
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        rpc.shutdown_timeout = TEST_DELAY
        await rpc.connect()
        await rpc.subscribe(request_handler=lambda x: x, queue=SUBSCRIPTION_QUEUE)

        rpc._tasks = [asyncio.sleep(small_delay) for i in range(number_of_tasks)]
        if pending:
            # a big delay task
            rpc._tasks.append(asyncio.sleep(big_delay))

        await asyncio.sleep(TEST_DELAY)
        future = asyncio.ensure_future(rpc.stop())
        await asyncio.sleep(TEST_DELAY)

        # if self._tasks contains a big delay task, the operation shouldn't be done
        if pending:
            await asyncio.sleep(big_delay)

        assert future.done()

    async def test_run(self):
        """
        Test for cabbagok.AsyncAmqpRpc.run(). This function requires
        termination of cabbagok.AsyncAmqpRpc.wait_connected() function
        """

        class FakeRunner:
            def __init__(self, delay_):
                self.delay = delay_

            async def run_server(self):
                pass

            async def wait_connected(self):
                await asyncio.sleep(self.delay)

        delay = TEST_DELAY
        delta = 0.1 * TEST_DELAY
        fake_runner = FakeRunner(delay)
        future = asyncio.ensure_future(cabbagok.AsyncAmqpRpc.run(fake_runner))

        await asyncio.sleep(delay + delta)

        # wait_connected() have been terminated and cabbagok.AsyncAmqpRpc.run() should also terminate
        assert future.done()

    async def test_run_server(self):
        """
        Test for cabbagok.AsyncAmqpRpc.run_server. The fuction shouldn't terminate until
        the variable self.keep_running is True.
        """

        class FakeProtocol:
            def __init__(self, delay):
                self.delay = delay

            async def wait_closed(self):
                await asyncio.sleep(self.delay)

        class FakeConnection:
            def __init__(self, delay):
                self.protocol = FakeProtocol(delay)
                self.disconnect_delay = delay

            async def disconnect(self):
                await asyncio.sleep(self.disconnect_delay)

        class FakeSelf:
            def __init__(self, delay):
                self.connection = FakeConnection(delay)
                self.keep_running = True
                self.start_subscriptions = [[random.random() for _ in range(5)] for _ in range(5)]

            async def subscribe(self, *params):
                pass

            async def connect(self):
                pass

        test_delay = TEST_DELAY
        delta = TEST_DELAY * 0.1
        fake_self = FakeSelf(test_delay)
        future = asyncio.ensure_future(cabbagok.AsyncAmqpRpc.run_server(fake_self))

        # While fake_self.keep_runnig the function shouldn't terminate
        await asyncio.sleep(3 * test_delay)
        assert not future.done()

        # After changing fake_self.keep_running the function should terminate
        # One delay is for self.connection.protocol.wait_closed,
        # another one is for self.connection.disconnect
        fake_self.keep_running = False
        await asyncio.sleep(2 * test_delay + delta)
        assert future.done()

    async def test_on_request(self, rpc):
        """
        Test for cabbagok.AsyncAmqpRpc._on_request
        It's checking that callback inside the function has been called
        """
        big_delay = 2 * TEST_DELAY

        async def run_delay(request):
            await asyncio.sleep(big_delay)

        delta = TEST_DELAY * 0.1
        rpc = await rpc
        await rpc.connect()
        asyncio.ensure_future(rpc._on_request(rpc.channel, b'', MockEnvelope(), MockProperties(), run_delay))

        # rpc._tasks shouldn't be empty
        await asyncio.sleep(TEST_DELAY)
        assert len(rpc._tasks) != 0

        # after request_handler argument of tested function is done, callback for clearing rpc._tasks should run
        await asyncio.sleep(TEST_DELAY + delta)
        assert len(rpc._tasks) == 0

    @pytest.mark.parametrize('shuffle', [True, False])
    async def test_shuffle(self, shuffle):
        """
        Test for cabbagok.AmqpConnection.cycle_hosts
        It checks correct work of parameter 'shuffle' and also changing the state of
        internal variable self.hosts during shuffling
        """

        # Generating random hosts
        hosts_length = 100
        hosts = [random.random() for _ in range(hosts_length)]

        # Creating a copy, and use its in constructor of AmqpConnection
        hosts_copy = hosts.copy()
        connection = cabbagok.AmqpConnection(hosts_copy)

        # Retrieving one period of hosts after cycle hosts
        retrieved_hosts_two_periods = list(islice(connection.cycle_hosts(shuffle), hosts_length * 2))
        retrieved_hosts = retrieved_hosts_two_periods[0:hosts_length]

        # Check for correct work of cycle
        assert retrieved_hosts + retrieved_hosts == retrieved_hosts_two_periods

        # cycle_hosts() always includes self.hosts
        assert retrieved_hosts == hosts_copy

        # Check for shuffling
        assert (hosts == retrieved_hosts) != shuffle



# -*- coding: utf-8 -*-
import itertools

from cabbagok.utils import ExponentialBackoff, FibonaccianBackoff


class TestExponentialBackoff:

    def test_double(self):
        backoff = ExponentialBackoff(1, 2, 2048)
        assert list(itertools.islice(backoff, 15)) == [
            1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 2048, 2048, 2048
        ]


class TestFibonaccianBackoff:

    def test_maxprecision(self):
        backoff = FibonaccianBackoff(1000)
        assert list(itertools.islice(backoff, 15)) == [
            1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610
        ]

    def test_1min(self):
        backoff = FibonaccianBackoff(60)
        assert list(itertools.islice(backoff, 15)) == [
            1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 60, 60, 60, 60, 60
        ]



# -*- coding: utf-8 -*-

from cabbagok import test_utils
from .amqp import AmqpConnection, AsyncAmqpRpc, ServiceUnavailableError

__all__ = ['ServiceUnavailableError', 'AmqpConnection', 'AsyncAmqpRpc', 'test_utils']

__version__ = '1.1.0'



# -*- coding: utf-8 -*-
import pytest

from cabbagok import AmqpConnection, ServiceUnavailableError
from cabbagok.test_utils import FakeAsyncAmqpRpc


def request_handler(request):
    return request


async def request_handler_async(request):
    return request


@pytest.mark.parametrize('handler', [request_handler, request_handler_async])
@pytest.mark.parametrize('body', ['text', {'type': 'json'}])
@pytest.mark.asyncio
async def test_server(body, handler):
    routing_key = 'test'
    connection = AmqpConnection()
    subscriptions = [(handler, routing_key)]
    rpc = FakeAsyncAmqpRpc(connection, subscriptions=subscriptions)
    await rpc.run()

    result = await rpc.fake_message(routing_key, body)
    assert result == body


@pytest.mark.asyncio
async def test_awaitable_response():
    connection = AmqpConnection()
    rpc = FakeAsyncAmqpRpc(connection)
    with pytest.raises(ValueError):
        await rpc.fake_message('invalid-queue', None)


@pytest.mark.parametrize('data', ['text', {'type': 'json'}])
@pytest.mark.parametrize('expected_response', ['text', {'type': 'json'}])
@pytest.mark.asyncio
async def test_client(expected_response, data):
    routing_key = 'test'
    connection = AmqpConnection()
    rpc = FakeAsyncAmqpRpc(connection)
    rpc.set_response(routing_key, expected_response)
    await rpc.connect()

    response = await rpc.send_rpc(routing_key, data)
    assert response == expected_response
    assert rpc.call_args == [(routing_key, data)]

    not_awaitable = await rpc.send_rpc(routing_key, data, await_response=False)
    assert not_awaitable is None
    assert rpc.call_args == [(routing_key, data), (routing_key, data)]


@pytest.mark.asyncio
async def test_client_no_response():
    connection = AmqpConnection()
    rpc = FakeAsyncAmqpRpc(connection)
    with pytest.raises(ServiceUnavailableError):
        await rpc.send_rpc('test', None)



# -*- coding: utf-8 -*-
import asyncio
import inspect
import logging
import random
import socket
import ssl as ssl_module
import uuid
from functools import partial
from itertools import cycle
from typing import Optional, Callable, Union, Awaitable, Mapping, Dict

from aioamqp.channel import Channel
from aioamqp.envelope import Envelope
from aioamqp.properties import Properties
from aioamqp.protocol import AmqpProtocol, CONNECTING, OPEN

from .utils import FibonaccianBackoff

logger = logging.getLogger('cabbagok')


class ServiceUnavailableError(Exception):
    """External service unavailable. """


class AmqpConnection:
    def __init__(self, hosts=None, username='guest', password='guest', virtualhost='/', loop=None, ssl=False):
        """
        :param hosts: iterable with tuples (host, port), default localhost:5672
        :param username: AMQP login, default guest
        :param password: AMQP password, default guest
        :param virtualhost: AMQP virtual host, default /
        :param loop: asyncio event loop, default current event loop
        :param ssl: bool or SSLContext, uses default ssl context if True, default False
        """
        self.loop = loop or asyncio.get_event_loop()
        self.username = username
        self.password = password
        self.virtualhost = virtualhost
        self.hosts = hosts if hosts is not None else [('localhost', 5672)]
        self._connection_cycle = self.cycle_hosts()
        self.transport = None
        self.protocol = None
        self.ssl = ssl

    async def channel(self):
        if self.protocol is not None:
            return await self.protocol.channel()

    def cycle_hosts(self, shuffle=False):
        if shuffle:
            random.shuffle(self.hosts)
        yield from cycle(self.hosts)

    async def connect(self):
        """Connect to AMQP broker. On failure this function will endlessly try reconnecting.
        Do nothing if already connected or connecting.
        """
        if self.protocol is not None and self.protocol.state in [CONNECTING, OPEN]:
            return

        delay = FibonaccianBackoff(limit=60.0)
        for host, port in self._connection_cycle:
            try:
                self.transport, self.protocol = await aioamqp_connect(
                    loop=self.loop,
                    host=host,
                    port=port,
                    login=self.username,
                    password=self.password,
                    virtualhost=self.virtualhost,
                    ssl=self.ssl
                )
            except OSError as e:
                # Connection-related errors are mostly represented by `ConnectionError`,
                # except for some like `socket.gaierror`, which fall into broader `OSError`
                next_delay = delay.next()
                logger.warning(f'failed to connect to {host}:{port}, error <{e.__class__.__name__}> {e}, '
                               f'retrying in {next_delay} seconds')
                await asyncio.sleep(next_delay)
            except Exception as e:
                logger.error(f'connection failed, not retrying: <{e.__class__.__name__}> {e}')
                raise
            else:
                logger.info(f'wait connection to {host}:{port}')
                break

    async def disconnect(self):
        if self.protocol is not None and self.protocol.state in [CONNECTING, OPEN]:
            await self.protocol.close()

    @property
    def is_connected(self):
        """Property, required for rpc to check readiness"""
        return bool(self.protocol) and self.protocol.state == OPEN


class AsyncAmqpRpc:
    def __init__(self, connection: AmqpConnection,
                 exchange_params: Mapping = None, queue_params: Mapping = None,
                 subscriptions=None, prefetch_count=1, raw=False, default_response_timeout=15.0,
                 shutdown_timeout=60.0, connection_delay: float = 0.1, callback_exchange=''):
        """
        All arguments are optional. If `request_handler` is not supplied or None, RPC works only in client mode.

        :param queue_params: options for creating queues, default durable and DLX
        :param exchange_params: options when creating exchanges, default durable and type topic
        :param subscriptions: list of tuples (handler, queue, exchange, routing_key, queue_params, exchange_params)
                Rightmost parameters are optional, you can specify only (handler, queue).
        :param raw: do not attempt decoding, use iff `request_handler` maps `bytes -> bytes`.
        :param prefetch_count: per-consumer prefetch message limit, default 1
        :param default_response_timeout: default timeout for awaiting response when sending remote calls
        :param shutdown_timeout: timeout for handlers to finish gracefully on shutdown
        """
        self.raw = raw
        self.queue_params = queue_params
        self.exchange_params = exchange_params
        self.start_subscriptions = subscriptions or []
        self.default_response_timeout = default_response_timeout
        self.shutdown_timeout = shutdown_timeout
        self.connection = connection
        self.prefetch_count = prefetch_count
        self.keep_running = True
        self.channel = None
        self.callback_queue = None
        self.callback_exchange = callback_exchange
        self._responses: Dict[str, asyncio.Future] = {}
        self._tasks = set()
        self._subscriptions = set()
        self.connection_delay = connection_delay

    async def connect(self):
        await self.connection.connect()
        self.channel = await self.connection.channel()

        result = await self.channel.queue_declare(exclusive=True)
        self.callback_queue = result['queue']

        # setup client
        if self.callback_exchange != '':
            # operation not permitted on default exchange
            await self.channel.exchange_declare(exchange_name=self.callback_exchange, type_name='topic', durable=True)
            await self.channel.queue_bind(
                queue_name=self.callback_queue,
                exchange_name=self.callback_exchange,
                routing_key=self.callback_queue)

        await self.channel.basic_consume(
            callback=self._on_response,
            queue_name=self.callback_queue,
        )

        logger.debug(f'listening on callback queue {result["queue"]}')

    # AMQP server implementation

    async def declare(self, queue: str, exchange: str = '', routing_key: str = None,
                      queue_params: Mapping = None, exchange_params: Mapping = None):
        """
        Set up necessary objects — exchange, queue, binding, QoS.

        :param queue: queue name
        :param exchange: exchange name, default '' (default AMQP exchange)
        :param routing_key: routing key, default same as `queue`
        :param queue_params: options for the queue, default durable and DLX
        :param exchange_params: options for the exchange, default durable and type topic
        """
        if routing_key is None:
            routing_key = queue
        if exchange_params is None:
            exchange_params = dict(type_name='topic', durable=True)
        if queue_params is None:
            queue_params = dict(durable=True, arguments={
                'x-dead-letter-exchange': 'DLX',
                'x-dead-letter-routing-key': 'dlx_rpc',
            })
        if exchange != '':
            # operation not permitted on default exchange
            await self.channel.exchange_declare(exchange_name=exchange, **exchange_params)
        result = await self.channel.queue_declare(queue_name=queue, **queue_params)
        queue = result['queue']  # in case queue name was generated by broker
        if exchange != '':
            # operation not permitted on default exchange
            await self.channel.queue_bind(
                queue_name=queue,
                exchange_name=exchange,
                routing_key=routing_key)
        await self.channel.basic_qos(
            prefetch_count=self.prefetch_count,
            prefetch_size=0,
            connection_global=False,
        )

    async def subscribe(self, request_handler: Union[
        Callable[[str], Optional[str]],
        Callable[[bytes], Optional[bytes]],
        Callable[[str], Awaitable[Optional[str]]],
        Callable[[bytes], Awaitable[Optional[bytes]]],
    ], queue: str, exchange: str = '', routing_key: str = None, add_to_start: bool = False) -> str:
        """
        Subscribe to a specific queue. Exchange and queue will be created if they do not exist.

        :param request_handler: request handler, can be a normal or coroutine function
                that maps either str->str or bytes->bytes. If `request_handler` returns None,
                it is taken to mean no response is needed.
        :param exchange: exchange name, default '' (default AMQP exchange)
        :param queue: queue name
        :param routing_key: routing key, default same as `queue`
        :param add_to_start: add to start_subscriptions
        :return: consumer_tag
        """
        if routing_key is None:
            routing_key = queue
        await self.declare(queue=queue, exchange=exchange, routing_key=routing_key,
                           queue_params=self.queue_params, exchange_params=self.exchange_params)
        result = await self.channel.basic_consume(
            callback=partial(self._on_request, request_handler=request_handler),
            queue_name=queue,
        )
        consumer_tag = result['consumer_tag']
        self._subscriptions.add(consumer_tag)
        self.start_subscriptions.append(
            (request_handler, queue, exchange, routing_key)
        )
        logger.debug(f'subscribed to queue {queue}, bound to exchange {exchange} with key {routing_key} '
                     f'(consumer tag {consumer_tag})')
        return consumer_tag

    async def unsubscribe(self, consumer_tag: str):
        """
        Stop consuming on a queue.

        :param consumer_tag: consumer tag returned by `subscribe()`
        """
        logger.debug(f'unsubscribed from a queue (consumer tag {consumer_tag})')
        await self.channel.basic_cancel(consumer_tag=consumer_tag)

    async def _on_request(self, channel, body, envelope, properties, request_handler):
        """Run handle_rpc() in background. """
        task = asyncio.ensure_future(self.handle_rpc(channel, body, envelope, properties, request_handler))
        self._tasks.add(task)
        task.add_done_callback(lambda fut: self._tasks.remove(fut))

    async def handle_rpc(self, channel: Channel, body: bytes, envelope: Envelope, properties: Properties,
                         request_handler):
        """Process request with handler and send response if needed. """
        try:
            data = body if self.raw else body.decode('utf-8')
            logger.debug(f'> handle_rpc: data {data}, routing_key {properties.reply_to}, '
                         f'correlation_id {properties.correlation_id}')
            response = request_handler(data)
            if inspect.isawaitable(response):
                response = await response
        except Exception as e:
            logger.error(f'handle_rpc. error <{e.__class__.__name__}> {e}, routing_key {properties.reply_to}, '
                         f'correlation_id {properties.correlation_id}')
            await channel.basic_client_nack(delivery_tag=envelope.delivery_tag)
        else:
            responding = properties.reply_to is not None and response is not None
            logger.debug(f'{"< " * responding}handle_rpc: responding? {responding}, routing_key {properties.reply_to}, '
                         f'correlation_id {properties.correlation_id}, result {response}')
            if responding:
                response_params = dict(
                    payload=response if self.raw else response.encode('utf-8'),
                    exchange_name='',
                    routing_key=properties.reply_to
                )

                if properties.correlation_id:
                    response_params['properties'] = {'correlation_id': properties.correlation_id}

                await channel.basic_publish(**response_params)

            if channel.is_open:
                await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)

    async def run_server(self):
        """Main routine for the server. """
        try:
            while self.keep_running:
                await self.connect()
                # TODO: reconnect to manual subscriptions on lost connection
                for params in self.start_subscriptions:
                    await self.subscribe(*params)
                await self.connection.protocol.wait_closed()
        finally:
            await self.connection.disconnect()

    async def run(self, app=None):
        """aiohttp-compatible on_startup coroutine. """
        asyncio.ensure_future(self.run_server())
        await self.wait_connected()
        logger.info('Waiting finished. Connected successfully.')

    async def stop(self, app=None):
        """aiohttp-compatible on_shutdown coroutine. """
        for consumer_tag in self._subscriptions:
            await self.unsubscribe(consumer_tag)
        if self._tasks:
            logger.info(f'waiting for {len(self._tasks)} task(s) to finish normally')
            done, pending = await asyncio.wait(self._tasks, timeout=self.shutdown_timeout)
            if pending:
                level = logger.warning
            else:
                level = logger.info
            level(f'{len(done)} task(s) finished, {len(pending)} task(s) did not finish in time')
        self.keep_running = False
        await self.connection.disconnect()

    # AMQP client implementation

    async def send_rpc(self, destination: str, data: Union[str, bytes], exchange: str = '', await_response=True,
                       timeout: float = None, correlation_id: str = None) -> Union[str, bytes, None]:
        """
        Execute a method on remote server. Sends `data` to `destination` routing key.

        If `await_response` is True, the call blocks coroutine until the result is returned or until `timeout` seconds
        passed (class default is used if None). AMQP correlation_id is set to `correlation_id` or new UUID if None.

        Raises `ServiceUnavailableError` on response timeout.
        """
        # TODO: retry on error?
        if isinstance(data, str):
            payload = data.encode('utf-8')
            raw = False
        else:
            payload = data
            raw = True
        properties = dict()
        if await_response:
            if correlation_id is None:
                correlation_id = str(uuid.uuid4())
            properties = {
                'reply_to': self.callback_queue,
                'correlation_id': correlation_id,
            }
        logger.debug(f'< send_rpc: destination {destination}, data {data}, '
                     f'awaiting? {await_response}, timeout {timeout}, properties {properties}')
        await self.channel.basic_publish(
            exchange_name=exchange,
            routing_key=destination,
            properties=properties,
            payload=payload,
        )

        if await_response:
            if timeout is None:
                timeout = self.default_response_timeout
            response = await self._await_response(correlation_id=correlation_id, timeout=timeout)
            if not raw:
                response = response.decode('utf-8')
            logger.debug(f'> send_rpc: response {response}')
            return response

    async def _await_response(self, correlation_id, timeout):
        """Wait for a response with given correlation id. Blocks current Task. """
        self._responses[correlation_id] = asyncio.Future()
        try:
            await asyncio.wait_for(self._responses[correlation_id], timeout=timeout)
            return self._responses[correlation_id].result()
        except asyncio.TimeoutError:
            logger.warning(f'request {correlation_id} timed out')
            raise ServiceUnavailableError('Request timed out') from None
        finally:
            self._responses.pop(correlation_id)

    async def _on_response(self, channel: Channel, body: bytes, envelope: Envelope, properties: Properties):
        """Set response result. Called by aioamqp on a message in callback queue. """
        if properties.correlation_id in self._responses:
            self._responses[properties.correlation_id].set_result(body)
        else:
            logger.warning(f'unexpected message with correlation_id {properties.correlation_id}.')

        if channel.is_open:
            await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)

    async def wait_connected(self):
        while not all([self.connection.is_connected, self.channel]):
            logger.debug(f'Waiting connection for {self.connection_delay}s...')
            await asyncio.sleep(self.connection_delay)


async def aioamqp_connect(host='localhost', port=None, login='guest', password='guest', virtualhost='/', ssl=False,
                          login_method='AMQPLAIN', insist=False, protocol_factory=AmqpProtocol, *, verify_ssl=True,
                          loop=None, timeout=None, **kwargs):
    """Convenient method to connect to an AMQP broker
        :param host:          the host to connect to
        :param port:          broker port
        :param login:         login
        :param password:      password
        :param virtualhost:   AMQP virtualhost to use for this connection
        :param ssl:           bool: Create an SSL connection instead of a plain unencrypted one
                              SSLContext: Set custom SSLContext object
        :param verify_ssl:    Verify server's SSL certificate (True by default)
        :param login_method:  AMQP auth method
        :param insist:        Insist on connecting to a server
        :param protocol_factory: Factory to use, if you need to subclass AmqpProtocol
        :param loop:          Set the event loop to use
        :param kwargs:        Arguments to be given to the protocol_factory instance
        :return:              a tuple (transport, protocol) of an AmqpProtocol instance
    """
    SSL_PORT = 5671
    DEFAULT_PORT = 5672

    if loop is None:
        loop = asyncio.get_event_loop()

    def factory(): return protocol_factory(loop=loop, **kwargs)

    create_connection_kwargs = {}

    if ssl:
        ssl_context = ssl_module.create_default_context() if isinstance(ssl, bool) else ssl
        if not verify_ssl:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl_module.CERT_NONE
        create_connection_kwargs['ssl'] = ssl_context

    if port is None:
        if ssl:
            port = SSL_PORT
        else:
            port = DEFAULT_PORT

    transport, protocol = await loop.create_connection(
        factory, host, port, **create_connection_kwargs
    )

    # these 2 flags *may* show up in sock.type. They are only available on linux
    # see https://bugs.python.org/issue21327
    nonblock = getattr(socket, 'SOCK_NONBLOCK', 0)
    cloexec = getattr(socket, 'SOCK_CLOEXEC', 0)
    sock = transport.get_extra_info('socket')
    if sock is not None and (sock.type & ~nonblock & ~cloexec) == socket.SOCK_STREAM:
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    try:
        await protocol.start_connection(host, port, login, password, virtualhost, ssl=ssl, login_method=login_method,
                                        insist=insist)
    except Exception:
        await protocol.wait_closed(timeout=timeout)
        raise

    return transport, protocol



# -*- coding: utf-8 -*-
import inspect

from cabbagok.amqp import AsyncAmqpRpc, ServiceUnavailableError


class FakeAsyncAmqpRpc(AsyncAmqpRpc):
    """
    Can be used instead of AsyncAmqpRpc class for tests.
    Allows to set expected responses for testing amqp rpc client and to send fake message for testing a server.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subscriptions = {}
        self.responses = {}
        self.call_args = []

    async def connect(self):
        pass

    def set_response(self, routing_key, data):
        self.responses[routing_key] = data

    async def send_rpc(self, destination, data, exchange='', await_response=True, timeout=None, correlation_id=None):
        self.call_args.append((destination, data))
        if not await_response:
            return

        response = self.responses.get(destination)

        if response is None:
            raise ServiceUnavailableError()

        return response

    async def run_server(self):
        for request_handler, queue, *_ in self.start_subscriptions:
            self.subscriptions[queue] = request_handler

    async def run(self, app=None):
        await self.run_server()

    async def fake_message(self, queue, data):
        if queue not in self.subscriptions:
            raise ValueError(f'invalid queue name {queue}')

        request_handler = self.subscriptions[queue]
        response = request_handler(data)

        if inspect.isawaitable(response):
            response = await response

        return response



# -*- coding: utf-8 -*-
import asyncio
from unittest.mock import patch

import pytest

import cabbagok
from tests.conftest import RANDOM_QUEUE, TEST_DESTINATION, RESPONSE_CORR_ID, MockEnvelope, MockProperties, DELIVERY_TAG

pytestmark = pytest.mark.asyncio


class TestSendRpc:
    """AsyncAmqpRpc.send_rpc"""

    @pytest.mark.parametrize('data, expected_payload', [
        ('', b''),
        (b'', b''),
        ('Test message body. Тест', b'Test message body. \xd0\xa2\xd0\xb5\xd1\x81\xd1\x82'),
        (b'Raw message \xff', b'Raw message \xff'),
    ])
    async def test_no_response(self, connection, data, expected_payload):
        """Check that data is correctly encoded (if needed) and sent."""
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        await rpc.send_rpc(destination=TEST_DESTINATION, data=data, await_response=False)
        rpc.channel.basic_publish.assert_called_once_with(
            exchange_name='', routing_key=TEST_DESTINATION, properties={}, payload=expected_payload)

    @pytest.mark.parametrize('data, sent_payload, received_payload, expected_result', [
        ('request data', b'request data', b'', ''),
        (b'raw request', b'raw request', b'', b''),
        ('request data', b'request data', b'Test. \xd0\xa2\xd0\xb5\xd1\x81\xd1\x82', 'Test. Тест'),
        (b'raw request', b'raw request', b'Raw message \xff', b'Raw message \xff'),
    ])
    async def test_await_response(self, connection, data, sent_payload, received_payload, expected_result):
        """Check that data returned by await_response is parsed and returned correctly."""
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        with patch('cabbagok.amqp.AsyncAmqpRpc._await_response', return_value=received_payload, autospec=True), \
                patch('cabbagok.amqp.uuid.uuid4', return_value=RESPONSE_CORR_ID):
            result = await rpc.send_rpc(destination=TEST_DESTINATION, data=data, await_response=True)
        assert result == expected_result
        rpc.channel.basic_publish.assert_called_once_with(
            exchange_name='', routing_key=TEST_DESTINATION, payload=sent_payload,
            properties={'reply_to': RANDOM_QUEUE, 'correlation_id': RESPONSE_CORR_ID})

    @pytest.mark.parametrize('correlation_id, expected', [
        ('my-correlation-id', 'my-correlation-id'),
        (None, RESPONSE_CORR_ID),  # generated by the library
    ])
    async def test_correlation_id(self, connection, correlation_id, expected):
        """Check that correlation id either matches the custom id or is generated from uuid."""
        rpc = cabbagok.AsyncAmqpRpc(connection=connection)
        await rpc.connect()
        with patch('cabbagok.amqp.AsyncAmqpRpc._await_response', return_value=b'response', autospec=True), \
                patch('cabbagok.amqp.uuid.uuid4', return_value=RESPONSE_CORR_ID):
            await rpc.send_rpc(destination=TEST_DESTINATION, data='payload', await_response=True,
                               correlation_id=correlation_id)
        rpc.channel.basic_publish.assert_called_once_with(
            exchange_name='', routing_key=TEST_DESTINATION, payload=b'payload',
            properties={'reply_to': RANDOM_QUEUE, 'correlation_id': expected})


class TestAwaitResponse:
    """AsyncAmqpRpc._await_response"""

    async def test_ok(self, rpc):
        rpc = await rpc
        # schedule awaiting response in another Task
        task = asyncio.ensure_future(rpc._await_response(correlation_id=RESPONSE_CORR_ID, timeout=10.0))
        # but it's not executing yet
        assert set(rpc._responses.keys()) == set()
        # let it run for a bit
        await asyncio.sleep(0)
        # check that it created a Future
        assert set(rpc._responses.keys()) == {RESPONSE_CORR_ID}
        # set Future result
        rpc._responses[RESPONSE_CORR_ID].set_result('task result')
        # let the Task run to completion
        await task
        assert task.done()
        assert task.result() == 'task result'
        # check that it cleaned up rpc._responses
        assert set(rpc._responses.keys()) == set()

    async def test_timeout(self, rpc):
        rpc = await rpc
        with pytest.raises(cabbagok.ServiceUnavailableError):
            await rpc._await_response(correlation_id=RESPONSE_CORR_ID, timeout=0)
        assert set(rpc._responses.keys()) == set()


class TestOnResponse:
    """AsyncAmqpRpc._on_response: aioamqp handler for messages in callback queue."""

    @pytest.mark.parametrize('body', [b'', b'Test message body. \xd0\xa2\xd0\xb5\xd1\x81\xd1\x82'])
    async def test_ok(self, rpc, body):
        rpc = await rpc
        rpc._responses[RESPONSE_CORR_ID] = asyncio.Future()
        await rpc._on_response(channel=rpc.channel, body=body, envelope=MockEnvelope(), properties=MockProperties())
        assert rpc._responses[RESPONSE_CORR_ID].done()
        assert rpc._responses[RESPONSE_CORR_ID].result() == body
        rpc.channel.basic_client_ack.assert_called_once_with(delivery_tag=DELIVERY_TAG)
        rpc.channel.basic_client_nack.assert_not_called()

    async def test_unexpected_tag(self, rpc):
        rpc = await rpc
        await rpc._on_response(channel=rpc.channel, body=b'resp', envelope=MockEnvelope(), properties=MockProperties())
        rpc.channel.basic_client_ack.assert_called_once_with(delivery_tag=DELIVERY_TAG)



# -*- coding: utf-8 -*-
import os
from setuptools import setup


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


def get_version():
    """Get version from the package without actually importing it."""
    init = read('cabbagok/__init__.py')
    for line in init.split('\n'):
        if line.startswith('__version__'):
            return eval(line.split('=')[1])


setup(
    name='cabbagok',
    version=get_version(),
    description='asyncio-based AMQP client and server for RPC.',
    packages=['cabbagok'],
    url='https://github.com/whiteapfel/cabbagok/',
    download_url='https://pypi.org/project/cabbagok/',
    install_requires=read('requirements.txt'),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    maintainer='WhiteApfel',
    maintainer_email='white@pfel.ru',
    license='Mozilla Public License 2.0',
    python_requires='>=3.11',
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)
